package cn.minbb.serialport;

import android.os.Handler;
import android.os.HandlerThread;
import android.util.Log;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;

import cn.minbb.serialport.app.DataType;
import cn.minbb.serialport.listener.OnDataProgressListener;
import cn.minbb.serialport.listener.OnOpenSerialPortListener;
import cn.minbb.serialport.listener.OnSerialPortDataListener;
import cn.minbb.serialport.thread.SerialPortReadThread;
import cn.minbb.serialport.utils.ByteUtil;
import cn.minbb.serialport.utils.LogUtil;
import cn.minbb.serialport.utils.PackageUtil;

/**
 * 串口管理器 - 单例模式  防止 反射 克隆 序列化 破坏
 */
public class SerialPortManager extends SerialPort implements Serializable, Cloneable {

    private static final String TAG = SerialPortManager.class.getSimpleName();
    private FileInputStream mFileInputStream;
    private FileOutputStream mFileOutputStream;
    private FileDescriptor mFd;
    private OnOpenSerialPortListener mOnOpenSerialPortListener;
    private OnSerialPortDataListener mOnSerialPortDataListener;
    private OnDataProgressListener onDataProgressListener;

    private HandlerThread mSendingHandlerThread;
    private SerialPortReadThread mSerialPortReadThread;

    private SerialPortStatus serialPortStatus = SerialPortStatus.CLOSE;

    // 报文长度
    private static int LENGTH = 64;
    // 超时时间
    private static int TIMEOUT = 5000;
    // 会话 ID
    private String sessionId = null;
    // 流水 ID
    private String serialId = null;
    // 字节缓冲区列表
    private List<byte[]> bytesBufferList = new ArrayList<>();

    // 默认是第一次创建
    private static volatile boolean isCreate = false;
    // 单例对象
    private static SerialPortManager serialPortManager = null;

    private SerialPortManager() {
        if (isCreate) {
            throw new RuntimeException("已然被实例化一次，不能在实例化");
        }
        isCreate = true;
    }

    /**
     * 线程安全 双重检查加锁
     */
    public static SerialPortManager getInstance() {
        if (null == serialPortManager) {
            synchronized (SerialPortManager.class) {
                if (null == serialPortManager) {
                    serialPortManager = new SerialPortManager();
                }
            }
        }
        return serialPortManager;
    }

    /**
     * 同步设置串口状态
     */
    private synchronized void setSerialPortStatus(SerialPortStatus serialPortStatus) {
        this.serialPortStatus = serialPortStatus;
    }

    /**
     * 同步获取当前串口状态
     */
    public synchronized SerialPortStatus getSerialPortStatus() {
        return this.serialPortStatus;
    }

    /**
     * 同步获取会话 ID
     */
    public synchronized String getSessionId() {
        return sessionId;
    }

    /**
     * 同步设置流水号
     */
    public synchronized void setSerialId(String serialId) {
        this.serialId = serialId;
    }

    /**
     * 同步获取流水号
     */
    public synchronized String getSerialId() {
        return serialId;
    }

    /**
     * 同步获取字节缓冲区列表
     */
    public synchronized List<byte[]> getBytesBufferList() {
        return bytesBufferList;
    }

    /**
     * 收到串口数据调用，同步锁
     *
     * @param bytes 收到的 Byte 数组
     */
    private synchronized void dataBytesReceived(byte[] bytes) {
        bytesBufferList.add(bytes);
    }

    /**
     * 清空字符缓冲区列表
     */
    private synchronized void dataBytesReceivedOver() {
        bytesBufferList.clear();
    }

    /**
     * 设置接收数据超时时间
     *
     * @param timeout 毫秒
     */
    public static void setTimeout(int timeout) {
        SerialPortManager.TIMEOUT = timeout;
    }

    /**
     * 打开串口
     * <p>
     * 串口设备
     * 波特率
     *
     * @return 打开是否成功返回会话 ID，失败返回 null
     */
    public String openDefaultSerialPort() {
        return openCustomSerialPort(new File("/dev/ttyS1"), 115200);
    }

    /**
     * 打开指定波特率的串口
     *
     * @param device   路径
     * @param baudRate 波特率
     * @return 打开是否成功返回会话 ID，失败返回 null
     */
    public String openCustomSerialPort(File device, int baudRate) {
        Log.i(TAG, "openSerialPort: " + String.format("打开串口 %s 波特率 %s", device.getPath(), baudRate));
        // 校验串口权限
        if (!device.canRead() || !device.canWrite()) {
            boolean chmod777 = chmod777(device);
            if (!chmod777) {
                Log.i(TAG, "openSerialPort: 没有读写权限");
                if (null != mOnOpenSerialPortListener) {
                    mOnOpenSerialPortListener.onFail(device, OnOpenSerialPortListener.Status.NO_READ_WRITE_PERMISSION);
                }
                return null;
            }
        }

        try {
            mFd = open(device.getAbsolutePath(), baudRate, 0);
            mFileInputStream = new FileInputStream(mFd);
            mFileOutputStream = new FileOutputStream(mFd);
            Log.i(TAG, "openSerialPort: 串口已经打开 " + mFd);
            if (null != mOnOpenSerialPortListener) {
                mOnOpenSerialPortListener.onSuccess(device);
            }
            // 开启发送消息的线程
            startSendThread();
            // 开启接收消息的线程
            startReadThread();
            setSerialPortStatus(SerialPortStatus.OPEN);
            return UUID.randomUUID().toString();
        } catch (Exception e) {
            e.printStackTrace();
            if (null != mOnOpenSerialPortListener) {
                mOnOpenSerialPortListener.onFail(device, OnOpenSerialPortListener.Status.OPEN_FAIL);
            }
        }
        return null;
    }

    /**
     * 关闭串口
     */
    public void closeSerialPort() {
        if (null != mFd) {
            close();
            mFd = null;
        }
        // 停止发送消息的线程
        stopSendThread();
        // 停止接收消息的线程
        stopReadThread();

        if (null != mFileInputStream) {
            try {
                mFileInputStream.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
            mFileInputStream = null;
        }

        if (null != mFileOutputStream) {
            try {
                mFileOutputStream.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
            mFileOutputStream = null;
        }

        mOnOpenSerialPortListener = null;
        mOnSerialPortDataListener = null;
        sessionId = null;
        setSerialPortStatus(SerialPortStatus.CLOSE);
    }

    /**
     * 添加打开串口监听
     *
     * @param listener listener
     * @return SerialPortManager
     */
    public SerialPortManager setOnOpenSerialPortListener(OnOpenSerialPortListener listener) {
        mOnOpenSerialPortListener = listener;
        return this;
    }

    /**
     * 添加数据通信监听
     *
     * @param listener listener
     * @return SerialPortManager
     */
    public SerialPortManager setOnSerialPortDataListener(OnSerialPortDataListener listener) {
        mOnSerialPortDataListener = listener;
        return this;
    }

    /**
     * 开启发送消息的线程
     */
    private void startSendThread() {
        // 开启发送消息的线程
        mSendingHandlerThread = new HandlerThread("mSendingHandlerThread");
        mSendingHandlerThread.start();
    }

    /**
     * 停止发送消息线程
     */
    private void stopSendThread() {
        if (null != mSendingHandlerThread) {
            mSendingHandlerThread.interrupt();
            mSendingHandlerThread.quit();
            mSendingHandlerThread = null;
        }
    }

    /**
     * 开启接收消息的线程
     */
    private void startReadThread() {
        mSerialPortReadThread = new SerialPortReadThread(mFileInputStream) {
            @Override
            public void onDataReceived(final byte[] bytes) {
                // 设置接收流水号
                if (null == getSerialId()) setSerialId(String.valueOf(System.currentTimeMillis()));
                // 收到 EOF(-1) 或者超时则认为数据包接收完毕
                // 串口空闲状态下收到数据，启动超时计时器，并修改串口状态为数据输入状态
                if (SerialPortStatus.OPEN == getSerialPortStatus()) {
                    setSerialPortStatus(SerialPortStatus.DATA_IN);
                    new Timer().schedule(new TimerTask() {
                        @Override
                        public void run() {
                            // 超时执行
                            if (SerialPortStatus.DATA_IN == getSerialPortStatus()) {
                                // 解包收到的数据
                                LogUtil.setAppendFile("接收超时");
                                dataNotify();
                            }
                        }
                    }, TIMEOUT);
                }
                // 遍历数据块，检测到报文结束符 EOF(-1) 则终止报文接收
                int eofIndex = -1;
                int blockSize = bytes.length;
                byte[] byteBlock = new byte[blockSize];
                for (int i = 0; i < blockSize; i++) {
                    if (bytes[i] == -1) {
                        eofIndex = i;
                        break;
                    } else {
                        byteBlock[i] = bytes[i];
                    }
                }
                if (eofIndex != -1) {
                    // 收到包尾，拼接包尾数据后通知
                    byte[] tailBytes = new byte[eofIndex];
                    System.arraycopy(byteBlock, 0, tailBytes, 0, eofIndex);
                    dataBytesReceived(tailBytes);
                    dataNotify();
                } else {
                    dataBytesReceived(byteBlock);
                }
            }
        };
        mSerialPortReadThread.start();
    }

    /**
     * 停止接收消息的线程
     */
    private void stopReadThread() {
        if (null != mSerialPortReadThread) {
            mSerialPortReadThread.release();
        }
    }

    /**
     * 发送字符数组数据
     *
     * @param sendBytes 字符数组
     */
    public void sendBytes(byte[] sendBytes, OnDataProgressListener onDataProgressListener) throws IOException {
        this.onDataProgressListener = onDataProgressListener;
        sendData(sendBytes, DataType.BYTES);
    }

    /**
     * 发送字符串数据（UTF-8编码）
     *
     * @param string 字符串
     */
    public void sendString(String string, OnDataProgressListener onDataProgressListener) throws IOException {
        this.onDataProgressListener = onDataProgressListener;
        sendData(string.getBytes(StandardCharsets.UTF_8), DataType.STRING);
    }

    /**
     * 发送文件数据
     *
     * @param file 文件
     */
    public void sendFile(File file, OnDataProgressListener onDataProgressListener) throws Exception {
        if (!file.exists()) throw new FileNotFoundException("文件不存在");
        if (file.isDirectory()) throw new RuntimeException("文件夹不可以发送");
        this.onDataProgressListener = onDataProgressListener;
        FileInputStream fis = new FileInputStream(file);
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        int len;
        byte[] buffer = new byte[64];
        while ((len = fis.read(buffer)) != -1) {
            byteArrayOutputStream.write(buffer, 0, len);
        }
        byte[] byteArray = byteArrayOutputStream.toByteArray();
        fis.close();
        byteArrayOutputStream.close();
        sendData(byteArray, DataType.FILE);
    }

    /**
     * 发送数据
     *
     * @param payload  数据负载
     * @param dataType 数据类型
     */
    private void sendData(byte[] payload, DataType dataType) throws IOException {
        if (null != mFd && null != mFileInputStream && null != mFileOutputStream) {
            setSerialPortStatus(SerialPortStatus.DATA_OUT);
            switch (dataType) {
                case BYTES:
                    LogUtil.setAppendFile("发送字符数组：" + Arrays.toString(payload));
                    mFileOutputStream.write(payload);
                    break;
                case STRING:
                    LogUtil.setAppendFile("发送字符：[" + new String(payload) + "] -> " + Arrays.toString(payload));
                    // 分包，发送数据
                    mFileOutputStream.write(payload);
                    break;
                case FILE:
                    int length = payload.length;        // 负载总长度
                    int blockSize = 64;                 // 每次发送块大小
                    int group = length / blockSize + 1; // 负载总发送组数
                    int tailSize = length % blockSize;  // 负载末组大小
                    for (int index = 1; index <= group; index++) {
                        byte[] byteBuffer = new byte[blockSize];
                        if (index < group) {
                            System.arraycopy(payload, (index - 1) * blockSize, byteBuffer, 0, blockSize);
                        } else {
                            // 最后一组 - 可能是非完整
                            byteBuffer = new byte[tailSize];
                            System.arraycopy(payload, (index - 1) * blockSize, byteBuffer, 0, tailSize);
                        }
                        if (byteBuffer.length != 0) {
                            // 发送数据包
                            mFileOutputStream.write(byteBuffer);
                            LogUtil.setAppendFile("发送文件第" + index + "组，共" + group + "组" + Arrays.toString(byteBuffer));
                            if (null != onDataProgressListener) {
                                onDataProgressListener.onProgress(index / (float) group, index, group, byteBuffer);
                            }
                        }
                    }
                    break;
            }
            setSerialPortStatus(SerialPortStatus.OPEN);
        } else {
            LogUtil.setAppendFile("数据流为空");
            throw new RuntimeException("数据流为空");
        }
    }

    /**
     * 数据回调方法，当串口接收数据超时或者收到数据结束符时候调用，用于通知应用程序收到的数据
     */
    private void dataNotify() {
        if (null != mOnSerialPortDataListener) {
            // TODO: 2020-04-16 解包
            String data = PackageUtil.analyze(getBytesBufferList(), getSerialId());
            mOnSerialPortDataListener.onDataReceived(null, data);
        } else {
            LogUtil.setAppendFile("串口数据监听器未定义");
            throw new RuntimeException("串口数据监听器未定义");
        }
        dataBytesReceivedOver();
        // 重置流水号
        setSerialId(null);
        // 串口状态设置为空闲
        setSerialPortStatus(SerialPortStatus.OPEN);
    }

    @Override
    protected Object clone() throws CloneNotSupportedException {
        return serialPortManager;
    }

    /**
     * 防止序列化破坏
     */
    private Object readResolve() {
        return serialPortManager;
    }
}
